/*
 *      Copyright (c) 2003 Michael T Pins.  All rights reserved.
 */

void            sort_articles(int);
void            no_sort_articles(void);
int             elim_articles(register article_number *, int);
