/*
 *      Copyright (c) 2003-2005 Michael T Pins.  All rights reserved.
 */

#ifndef NOV
void            nn_exit(int);
void            clean_group(group_header *);
group_header   *add_new_group(char *);
void            write_error(void);
#endif
