/*
 *      Copyright (c) 2003 Michael T Pins.  All rights reserved.
 */

void            uud_start(char *);
void            uud_end(void);
int             uudecode(register article_header *, FILE *);
