/*
 *      Copyright (c) 2003 Michael T Pins.  All rights reserved.
 */

int             expand_file_name(char *, char *, int);
int             file_completion(char *, int);
void            fcancel(article_header *);
int             folder_menu(char *, int);
