/*
 *      Copyright (c) 2003 Michael T Pins.  All rights reserved.
 */

void            init_execute(void);
int             execute(char *, char **, int);
int             shell_escape(void);
int             run_shell(char *, int, int);
int             suspend_nn(void);
