/*
 *      Copyright (c) 2003 Michael T Pins.  All rights reserved.
 */

void            init_message(char *,...);
void            visit_init_file(int, char *);
int             alt_completion(char *, int);
int             parse_command(char *, int, FILE *);
void            display_help(char *);
